"""Typing system for places (and transitions guards). Types are
contraint checkers to verify whether a value is in the type or not.
For instance:

>>> 5 in tInteger
True
>>> 4.3 in tInteger
False

The typechecking of several values is possible using the type as a
function. It fails if one value is not in the type.

>>> tInteger(5, 4, 6)
True
>>> tInteger(5, 4, 6.0)
False

Types can be composed in order to build more complexe types. For
example:

>>> 8 in (tInteger & ~Range(1, 5))
True
>>> 3 in (tInteger & ~Range(1, 5))
False

The various compositions are the same as sets operations, plus the
complement: `&`, `|`, `-`, `^` and `~` (complement).

Various types are predefined (like `tInteger`) the others can be
constructed using the various classes in the module. Prefefined types
are:

  * `tAll`: any value is in the type
  * `tNothing`: empty type
  * `tString`: string values
  * `tList`: lists values
  * `tInteger`: integer values
  * `tNatural`: non-negative integers
  * `tPositive`: strictly positive integers
  * `tFloat`: float values
  * `tNumber`: integers or float values
  * `tDict`: Python's `dict` values
  * `tNone`: the single value `None`
  * `tBoolean`: `True` or `False`
  * `tTuple`: tuple values
  * `tPair`: tuples of length two

Types with finitely many elements can be iterated over:

>>> list(sorted(iter(CrossProduct(Range(0, 10, 2), tBoolean) & ~OneOf((4, True), (6, False)))))
[(0, False), (0, True), (2, False), (2, True), (4, False), (6, True), (8, False), (8, True)]
>>> list(iter(OneOf(1, 2, 3)))
[1, 2, 3]
>>> iter(tInteger)
Traceback (most recent call last):
...
TypeError: ... not iterable
>>> iter(~OneOf(1, 2, 3))
Traceback (most recent call last):
...
TypeError: ... not iterable
"""

import inspect, sys
from snakes.compat import *
from snakes.data import cross
from snakes.pnml import Tree

def _iterable (obj, *types) :
    for t in types :
        try :
            iter(t)
        except :
            def __iterable__ () :
                raise ValueError("iteration over non-sequence")
            obj.__iterable__ = __iterable__
            break

class Type (object) :
    """Base class for all types. Implements operations `&`, `|`, `-`,
    `^` and `~` to build new types. Also implement the typechecking of
    several values. All the subclasses should implement the method
    `__contains__` to typecheck a single object.

    This class is abstract and should not be instantiated.
    """
    # apidoc skip
    def __init__ (self) :
        """Abstract method

        >>> Type()
        Traceback (most recent call last):
         ...
        NotImplementedError: abstract class

        @raise NotImplementedError: when called
        """
        raise NotImplementedError("abstract class")
    # apidoc skip
    def __eq__ (self, other) :
        return (self.__class__ == other.__class__
                and self.__dict__ == other.__dict__)
    # apidoc skip
    def __hash__ (self) :
        return hash(repr(self))
    def __and__ (self, other) :
        """Intersection type.

        >>> Instance(int) & Greater(0)
        (Instance(int) & Greater(0))

        @param other: the other type in the intersection
        @type other: `Type`
        @return: the intersection of both types
        @rtype: `Type`
        """
        if other is self :
            return self
        else :
            return _And(self, other)
    def __or__ (self, other) :
        """Union type.

        >>> Instance(int) | Instance(bool)
        (Instance(int) | Instance(bool))

        @param other: the other type in the union
        @type other: `Type`
        @return: the union of both types
        @rtype: `Type`
        """
        if other is self :
            return self
        else :
            return _Or(self, other)
    def __sub__ (self, other) :
        """Substraction type.

        >>> Instance(int) - OneOf([0, 1, 2])
        (Instance(int) - OneOf([0, 1, 2]))

        @param other: the other type in the substraction
        @type other: `Type`
        @return: the type `self` minus the type `other`
        @rtype: `Type`
        """
        if other is self :
            return tNothing
        else :
            return _Sub(self, other)
    def __xor__ (self, other) :
        """Disjoint union type.

        >>> Greater(0) ^ Instance(float)
        (Greater(0) ^ Instance(float))

        @param other: the other type in the disjoint union
        @type other: `Type`
        @return: the disjoint union of both types
        @rtype: `Type`
        """
        if other is self :
            return tNothing
        else :
            return _Xor(self, other)
    def __invert__ (self) :
        """Complementary type.

        >>> ~ Instance(int)
        (~Instance(int))

        @return: the complementary type
        @rtype: `Type`
        """
        return _Invert(self)
    # apidoc skip
    def __iterable__ (self) :
        """Called to test if a type is iterable

        Should be replaced in subclasses that are not iterable

        @raise ValueError: if not iterable
        """
        pass
    def __call__ (self, *values) :
        """Typecheck values.

        >>> Instance(int)(3, 4, 5)
        True
        >>> Instance(int)(3, 4, 5.0)
        False

        @param values: values that have to be checked
        @type values: `object`
        @return: `True` if all the values are in the types, `False`
            otherwise
        @rtype: `bool`
        """
        for v in values :
            if v not in self :
                return False
        return True
    __pnmltag__ = "type"
    _typemap = None
    # apidoc skip
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Load a `Type` from a PNML tree

        Uses the attribute `__pnmltype__` to know which type
        corresponds to a tag "<type domain='xxx'>"

        >>> s = List(tNatural | tBoolean).__pnmldump__()
        >>> Type.__pnmlload__(s)
        Collection(Instance(list),
                   ((Instance(int) & GreaterOrEqual(0))
                    | OneOf(True, False)))

        @param tree: the PNML tree to load
        @type tree: `snakes.pnml.Tree`
        @return: the loaded type
        @rtype: `Type`
        """
        if cls._typemap is None :
            cls._typemap = {}
            for n, c in inspect.getmembers(sys.modules[cls.__module__],
                                           inspect.isclass) :
                try :
                    cls._typemap[c.__pnmltype__] = c
                except AttributeError :
                    pass
        return cls._typemap[tree["domain"]].__pnmlload__(tree)

class _BinaryType (Type) :
    """A type build from two other ones

    This class allows to factorize the PNML related code for various
    binary types.
    """
    def __pnmldump__ (self) :
        return Tree(self.__pnmltag__, None,
                    Tree("left", None, Tree.from_obj(self._left)),
                    Tree("right", None, Tree.from_obj(self._right)),
                    domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        return cls(tree.child("left").child().to_obj(),
                   tree.child("right").child().to_obj())

class _And (_BinaryType) :
    "Intersection of two types"
    __pnmltype__ = "intersection"
    def __init__ (self, left, right) :
        self._left = left
        self._right = right
        _iterable(self, left)
    def __repr__ (self) :
        return "(%s & %s)" % (repr(self._left), repr(self._right))
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        return (value in self._left) and (value in self._right)
    def __iter__ (self) :
        self.__iterable__()
        for value in self._left :
            if value in self._right :
                yield value

class _Or (_BinaryType) :
    "Union of two types"
    __pnmltype__ = "union"
    def __init__ (self, left, right) :
        self._left = left
        self._right = right
        _iterable(self, left, right)
    def __repr__ (self) :
        return "(%s | %s)" % (repr(self._left), repr(self._right))
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        return (value in self._left) or (value in self._right)
    def __iter__ (self) :
        self.__iterable__()
        for value in (self._left & self._right) :
            yield value
        for value in (self._left ^ self._right) :
            yield value

class _Sub (_BinaryType) :
    "Subtyping by difference"
    __pnmltype__ = "difference"
    def __init__ (self, left, right)  :
        self._left = left
        self._right = right
        _iterable(self, left)
    def __repr__ (self) :
        return "(%s - %s)" % (repr(self._left), repr(self._right))
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        return (value in self._left) and (value not in self._right)
    def __iter__ (self) :
        self.__iterable__()
        for value in self._left :
            if value not in self._right :
                yield value

class _Xor (_BinaryType) :
    "Exclusive union of two types"
    __pnmltype__ = "xor"
    def __init__ (self, left, right) :
        self._left = left
        self._right = right
        _iterable(self, left, right)
    def __repr__ (self) :
        return "(%s ^ %s)" % (repr(self._left), repr(self._right))
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        if value in self._left :
            return value not in self._right
        else :
            return value in self._right
    def __iter__ (self) :
        self.__iterable__()
        for value in self._left :
            if value not in self._right :
                yield value
        for value in self._right :
            if value not in self._left :
                yield value

class _Invert (Type) :
    "Complement of a type"
    def __init__ (self, base) :
        self._base = base
        _iterable(self, None)
    def __repr__ (self) :
        return "(~%s)" % repr(self._base)
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        return (value not in self._base)
    __pnmltype__ = "complement"
    def __pnmldump__ (self) :
        return Tree(self.__pnmltag__, None,
                    Tree.from_obj(self._base),
                    domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        return cls(tree.child().to_obj())

class _All (Type) :
    "A type allowing for any value"
    def __init__ (self) :
        pass
    def __and__ (self, other) :
        return other
    def __or__ (self, other) :
        return self
    def __sub__ (self, other) :
        return ~other
    def __xor__ (self, other) :
        return ~other
    def __invert__ (self) :
        return tNothing
    def __repr__ (self) :
        return "tAll"
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: `object`
        @return: `True`
        @rtype: `bool`
        """
        return True
    def __call__ (self, *values) :
        """Typecheck values.

        @param values: values that have to be checked
        @type values: `objet`
        @return: `True`
        """
        return True
    __pnmltype__ = "universal"
    def __pnmldump__ (self) :
        """
        >>> tAll.__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="universal"/>
        </pnml>
        """
        return Tree(self.__pnmltag__, None, domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        """
        >>> Type.__pnmlload__(tAll.__pnmldump__())
        tAll
        """
        return cls()

class _Nothing (Type) :
    "A types with no value"
    def __init__ (self) :
        pass
    def __and__ (self, other) :
        return self
    def __or__ (self, other) :
        return other
    def __sub__ (self, other) :
        return self
    def __xor__ (self, other) :
        return other
    def __invert__ (self) :
        return tAll
    def __repr__ (self) :
        return "tNothing"
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: `object`
        @return: `False`
        @rtype: `bool`
        """
        return False
    def __call__ (self, *values) :
        """Typecheck values.

        @param values: values that have to be checked
        @type values: `objet`
        @return: `False`
        """
        return False
    def __iter__ (self) :
        pass
    __pnmltype__ = "empty"
    def __pnmldump__ (self) :
        return Tree(self.__pnmltag__, None, domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        return cls()

class Instance (Type) :
    """A type whose values are all instances of one class.

    >>> [1, 2] in Instance(list)
    True
    >>> (1, 2) in Instance(list)
    False
    """
    def __init__ (self, _class) :
        """Initialize the type

        >>> Instance(int)
        Instance(int)

        @param _class: the class of instance
        @type _class: `class`
        @return: initialized object
        @rtype: `Instance`
        """
        self._class = _class
    # apidoc stop
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        >>> 5 in Instance(int)
        True
        >>> 5.0 in Instance(int)
        False

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        return isinstance(value, self._class)
    def __repr__ (self) :
        """String representation of the type, suitable for `eval`

        >>> repr(Instance(str))
        'Instance(str)'

        @return: precise string representation
        @rtype: `str`
        """
        return "Instance(%s)" % self._class.__name__
    __pnmltype__ = "instance"
    def __pnmldump__ (self) :
        """Dump a type to a PNML tree

        >>> Instance(int).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="instance">
          <object name="int" type="class"/>
         </type>
        </pnml>

        @return: the PNML representation of the type
        @rtype: `str`
        """
        return Tree(self.__pnmltag__, None,
                    Tree.from_obj(self._class),
                    domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Builds a type from its PNML representation

        >>> t = Instance(int).__pnmldump__()
        >>> Instance.__pnmlload__(t)
        Instance(int)

        @param tree: the PNML tree to load
        @type tree: `snakes.pnml.Tree`
        @return: the loaded type
        @rtype: `Instance`
        """
        return cls(tree.child().to_obj())

def _full_name (fun) :
    if fun.__module__ is None :
        funname = fun.__name__
        for modname in sys.modules :
            try :
                if sys.modules[modname].__dict__.get(funname, None) is fun :
                    return ".".join([modname, funname])
            except :
                pass
        return funname
    else :
        return ".".join([fun.__module__, fun.__name__])

class TypeCheck (Type) :
    """A type whose values are accepted by a given function.

    >>> def odd (val) :
    ...     return type(val) is int and (val % 2) == 1
    >>> 3 in TypeCheck(odd)
    True
    >>> 4 in TypeCheck(odd)
    False
    >>> 3.0 in TypeCheck(odd)
    False
    """
    def __init__ (self, checker, iterate=None) :
        """Initialize the type, optionally, a function to iterate over
        the elements may be provided.

        >>> import operator
        >>> TypeCheck(operator.truth)
        TypeCheck(...truth)
        >>> list(TypeCheck(operator.truth))
        Traceback (most recent call last):
          ...
        ValueError: type not iterable
        >>> def true_values () : # enumerates only choosen values
        ...     yield True
        ...     yield 42
        ...     yield '42'
        ...     yield [42]
        ...     yield (42,)
        ...     yield {True: 42}
        >>> TypeCheck(operator.truth, true_values)
        TypeCheck(...truth, snakes.typing.true_values)
        >>> list(TypeCheck(operator.truth, true_values))
        [True, 42, '42', [42], (42,), {True: 42}]

        @param checker: a function that checks one value and returns
            `True` if it is in te type and `False` otherwise
        @type checker: `callable`
        @param iterate: `None` or an iterator over the values of the
            type
        @type iterate: `generator`
        """
        self._check = checker
        self._iterate = iterate
    # apidoc stop
    def __iter__ (self) :
        """
        >>> def odd (val) :
        ...     return type(val) is int and (val % 2) == 1
        >>> i = iter(TypeCheck(odd))
        Traceback (most recent call last):
          ...
        ValueError: type not iterable
        >>> def odd_iter () :
        ...     i = 1
        ...     while True :
        ...         yield i
        ...         yield -i
        ...         i += 2
        >>> i = iter(TypeCheck(odd, odd_iter))
        >>> next(i), next(i), next(i)
        (1, -1, 3)
        """
        try :
            return iter(self._iterate())
        except TypeError :
            raise ValueError("type not iterable")
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        >>> def odd (val) :
        ...     return type(val) is int and (val % 2) == 1
        >>> 3 in TypeCheck(odd)
        True
        >>> 4 in TypeCheck(odd)
        False
        >>> 3.0 in TypeCheck(odd)
        False

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        return self._check(value)
    def __repr__ (self) :
        """
        >>> def odd (val) :
        ...     return type(val) is int and (val % 2) == 1
        >>> repr(TypeCheck(odd))
        'TypeCheck(snakes.typing.odd)'
        >>> def odd_iter () :
        ...     i = 1
        ...     while True :
        ...         yield i
        ...         yield -i
        ...         i += 2
        >>> repr(TypeCheck(odd, odd_iter))
        'TypeCheck(snakes.typing.odd, snakes.typing.odd_iter)'
        """
        if self._iterate is None :
            return "%s(%s)" % (self.__class__.__name__,
                               _full_name(self._check))
        else :
            return "%s(%s, %s)" % (self.__class__.__name__,
                                   _full_name(self._check),
                                   _full_name(self._iterate))
    __pnmltype__ = "checker"
    def __pnmldump__ (self) :
        """Dump type to a PNML tree

        >>> import operator
        >>> TypeCheck(operator.truth).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="checker">
          <checker>
           ...
          </checker>
          <iterator>
           <object type="NoneType"/>
          </iterator>
         </type>
        </pnml>
        >>> def true_values () : # enumerates only choosen values
        ...     yield True
        ...     yield 42
        ...     yield '42'
        ...     yield [42]
        ...     yield (42,)
        ...     yield {True: 42}
        >>> TypeCheck(operator.truth, true_values).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="checker">
          <checker>
           ...
          </checker>
          <iterator>
           <object name="snakes.typing.true_values" type="function"/>
          </iterator>
         </type>
        </pnml>

        Note that this last example would not work as `true_value` as
        been defined inside a docstring. In order to allow it to be
        re-loaded from PNML, it should be defined at module level.

        @return: the type serialized to PNML
        @rtype: `snakes.pnml.Tree`
        """
        return Tree(self.__pnmltag__, None,
                    Tree("checker", None, Tree.from_obj(self._check)),
                    Tree("iterator", None, Tree.from_obj(self._iterate)),
                    domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Build type from a PNML tree

        >>> import operator
        >>> TypeCheck.__pnmlload__(TypeCheck(operator.truth).__pnmldump__())
        TypeCheck(....truth)
        >>> def true_values () : # enumerates only choosen values
        ...     yield True
        ...     yield 42
        ...     yield '42'
        ...     yield [42]
        ...     yield (42,)
        ...     yield {True: 42}

        @param tree: the PNML tree to load
        @type tree: `snakes.pnml.Tree`
        @return: the loaded type
        @rtype: `TypeChecker`
        """
        return cls(tree.child("checker").child().to_obj(),
                   tree.child("iterator").child().to_obj())

class OneOf (Type) :
    """A type whose values are explicitely enumerated.

    >>> 3 in OneOf(1, 2, 3, 4, 5)
    True
    >>> 0 in OneOf(1, 2, 3, 4, 5)
    False
    """
    # apidoc stop
    def __init__ (self, *values) :
        """
        @param values: the enumeration of the values in the type
        @type values: `object`
        """
        self._values = values
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        >>> 3 in OneOf(1, 2, 3, 4, 5)
        True
        >>> 0 in OneOf(1, 2, 3, 4, 5)
        False

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        return value in self._values
    def __repr__ (self) :
        """
        >>> repr(OneOf(1, 2, 3, 4, 5))
        'OneOf(1, 2, 3, 4, 5)'
        """
        return "OneOf(%s)" % ", ".join([repr(val) for val in self._values])
    def __iter__ (self) :
        """
        >>> list(iter(OneOf(1, 2, 3, 4, 5)))
        [1, 2, 3, 4, 5]
        """
        return iter(self._values)
    __pnmltype__ = "enum"
    def __pnmldump__ (self) :
        """Dump type to its PNML representation

        >>> OneOf(1, 2, 3, 4, 5).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="enum">
          <object type="int">
           1
          </object>
          <object type="int">
           2
          </object>
          <object type="int">
           3
          </object>
          <object type="int">
           4
          </object>
          <object type="int">
           5
          </object>
         </type>
        </pnml>

        @return: PNML representation of the type
        @rtype: `snakes.pnml.Tree`
        """
        return Tree(self.__pnmltag__, None,
                    *(Tree.from_obj(val) for val in self._values),
                    **dict(domain=self.__pnmltype__))
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Build type from its PNML representation

        >>> OneOf.__pnmlload__(OneOf(1, 2, 3, 4, 5).__pnmldump__())
        OneOf(1, 2, 3, 4, 5)

        @param tree: PNML tree to load
        @type tree: `snakes.pnml.Tree`
        @return: loaded type
        @rtype: `OneOf`
        """
        return cls(*(child.to_obj() for child in tree.children))

class Collection (Type) :
    """A type whose values are a given container, holding items of a
    given type and ranging in a given interval.

    >>> [0, 1.1, 2, 3.3, 4] in Collection(Instance(list), tNumber, 3, 10)
    True
    >>> [0, 1.1] in Collection(Instance(list), tNumber, 3, 10) #too short
    False
    >>> [0, '1.1', 2, 3.3, 4] in Collection(Instance(list), tNumber, 3, 10) #wrong item
    False
    """
    def __init__ (self, collection, items, min=None, max=None) :
        """Initialise the type

        >>> Collection(Instance(list), tNumber, 3, 10)
        Collection(Instance(list), (Instance(int) | Instance(float)), min=3, max=10)

        @param collection: the collection type
        @type collection: `Type`
        @param items: the type of the items
        @type items: `Type`
        @param min: the smallest allowed value
        @type min: `object`
        @param max: the greatest allowed value
        @type max: `object`
        """
        self._collection = collection
        self._class = collection._class
        self._items = items
        self._max = max
        self._min = min
    # apidoc stop
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        >>> [0, 1.1, 2, 3.3, 4] in Collection(Instance(list), tNumber, 3, 10)
        True
        >>> [0, 1.1] in Collection(Instance(list), tNumber, 3, 10) #too short
        False
        >>> [0, '1.1', 2, 3.3, 4] in Collection(Instance(list), tNumber, 3, 10) #wrong item
        False

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        if value not in self._collection :
            return False
        try :
            len(value)
            iter(value)
        except TypeError :
            return False
        if (self._min is not None) and (len(value) < self._min) :
            return False
        if (self._max is not None) and (len(value) > self._max) :
            return False
        for item in value :
            if item not in self._items :
                return False
        return True
    def __repr__ (self) :
        """
        >>> repr(Collection(Instance(list), tNumber, 3, 10))
        'Collection(Instance(list), (Instance(int) | Instance(float)), min=3, max=10)'
        """
        if (self._min is None) and (self._max is None) :
            return "Collection(%s, %s)" % (repr(self._collection),
                                           repr(self._items))
        elif self._min is None :
            return "Collection(%s, %s, max=%s)" % (repr(self._collection),
                                                   repr(self._items),
                                                   repr(self._max))
        elif self._max is None :
            return "Collection(%s, %s, min=%s)" % (repr(self._collection),
                                                   repr(self._items),
                                                   repr(self._min))
        else :
            return "Collection(%s, %s, min=%s, max=%s)" % (repr(self._collection),
                                                           repr(self._items),
                                                           repr(self._min),
                                                           repr(self._max))
    __pnmltype__ = "collection"
    def __pnmldump__ (self) :
        """Dump type to a PNML tree

        >>> Collection(Instance(list), tNumber, 3, 10).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="collection">
          <container>
           <type domain="instance">
            <object name="list" type="class"/>
           </type>
          </container>
          <items>
           <type domain="union">
            <left>
             <type domain="instance">
              <object name="int" type="class"/>
             </type>
            </left>
            <right>
             <type domain="instance">
              <object name="float" type="class"/>
             </type>
            </right>
           </type>
          </items>
          <min>
           <object type="int">
            3
           </object>
          </min>
          <max>
           <object type="int">
            10
           </object>
          </max>
         </type>
        </pnml>

        @return: the PNML representation of the type
        @rtype: `snakes.pnml.Tree`
        """
        return Tree(self.__pnmltag__, None,
                    Tree("container", None, Tree.from_obj(self._collection)),
                    Tree("items", None, Tree.from_obj(self._items)),
                    Tree("min", None, Tree.from_obj(self._min)),
                    Tree("max", None, Tree.from_obj(self._max)),
                    domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Load type from its PNML representation

        >>> t = Collection(Instance(list), tNumber, 3, 10).__pnmldump__()
        >>> Collection.__pnmlload__(t)
        Collection(Instance(list), (Instance(int) | Instance(float)),
                   min=3, max=10)

        @param tree: the PNML tree to load
        @type tree: `snakes.pnml.Tree`
        @return: the loaded type
        @rtype: `Collection`
        """
        return cls(tree.child("container").child().to_obj(),
                   tree.child("items").child().to_obj(),
                   tree.child("min").child().to_obj(),
                   tree.child("max").child().to_obj())

def List (items, min=None, max=None) :
    """Shorthand for instantiating `Collection`

    >>> List(tNumber, min=3, max=10)
    Collection(Instance(list), (Instance(int) | Instance(float)), min=3, max=10)

    @param items: the type of the elements in the collection
    @type items: `Type`
    @param min: the minimum number of elements in the collection or
        `None`
    @type min: `int`
    @param max: the maximum number of elements in the collection or
        `None`
    @type max: `int`
    @return: a type that checks the given constraints
    @rtype: `Collection`
    """
    return Collection(Instance(list), items, min, max)

def Tuple (items, min=None, max=None) :
    """Shorthand for instantiating `Collection`

    >>> Tuple(tNumber, min=3, max=10)
    Collection(Instance(tuple), (Instance(int) | Instance(float)), min=3, max=10)

    @param items: the type of the elements in the collection
    @type items: `Type`
    @param min: the minimum number of elements in the collection or
        `None`
    @type min: `int`
    @param max: the maximum number of elements in the collection or
        `None`
    @type max: `int`
    @return: a type that checks the given constraints
    @rtype: `Collection`
    """
    return Collection(Instance(tuple), items, min, max)

def Set (items, min=None, max=None) :
    """Shorthand for instantiating `Collection`

    >>> Set(tNumber, min=3, max=10)
    Collection(Instance(set), (Instance(int) | Instance(float)), min=3, max=10)

    @param items: the type of the elements in the collection
    @type items: `Type`
    @param min: the minimum number of elements in the collection or
        `None`
    @type min: `int`
    @param max: the maximum number of elements in the collection or
        `None`
    @type max: `int`
    @return: a type that checks the given constraints
    @rtype: `Collection`
    """
    return Collection(Instance(set), items, min, max)

class Mapping (Type) :
    """A type whose values are mapping (eg, `dict`)

    >>> {'Yes': True, 'No': False} in Mapping(tString, tAll)
    True
    >>> {True: 1, False: 0} in Mapping(tString, tAll)
    False
    """
    def __init__ (self, keys, items, _dict=Instance(dict)) :
        """Initialise a mapping type

        >>> Mapping(tInteger, tFloat)
        Mapping(Instance(int), Instance(float), Instance(dict))
        >>> from snakes.data import hdict
        >>> Mapping(tInteger, tFloat, Instance(hdict))
        Mapping(Instance(int), Instance(float), Instance(hdict))

        @param keys: the type for the keys
        @type keys: `Type`
        @param items: the type for the items
        @type items: `Type`
        @param _dict: the class that mapping must be instances of
        @type _dict: `dict`
        """
        self._keys = keys
        self._items = items
        self._dict = _dict
    # apidoc stop
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        >>> {'Yes': True, 'No': False} in Mapping(tString, tAll)
        True
        >>> {True: 1, False: 0} in Mapping(tString, tAll)
        False

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        if not self._dict(value) :
            return False
        for key, item in value.items() :
            if key not in self._keys :
                return False
            if item not in self._items :
                return True
        return True
    def __repr__ (self) :
        """Return a string representation of the type suitable for `eval`

        >>> repr(Mapping(tString, tAll))
        'Mapping(Instance(str), tAll, Instance(dict))'

        @return: precise string representation
        @rtype: `str`
        """
        return "Mapping(%s, %s, %s)" % (repr(self._keys),
                                        repr(self._items),
                                        repr(self._dict))
    __pnmltype__ = "mapping"
    def __pnmldump__ (self) :
        """Dump type to a PNML tree

        >>> from snakes.hashables import hdict
        >>> Mapping(tString, tAll, Instance(hdict)).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="mapping">
          <keys>
           <type domain="instance">
            <object name="str" type="class"/>
           </type>
          </keys>
          <items>
           <type domain="universal"/>
          </items>
          <container>
           <type domain="instance">
            <object name="snakes.hashables.hdict" type="class"/>
           </type>
          </container>
         </type>
        </pnml>

        @return: PNML representation of the type
        @rtype: `snakes.pnml.Tree`
        """
        return Tree(self.__pnmltag__, None,
                    Tree("keys", None, Tree.from_obj(self._keys)),
                    Tree("items", None, Tree.from_obj(self._items)),
                    Tree("container", None, Tree.from_obj(self._dict)),
                    domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Load type from its PNML representation

        >>> from snakes.hashables import hdict
        >>> t = Mapping(tString, tAll, Instance(hdict)).__pnmldump__()
        >>> Mapping.__pnmlload__(t)
        Mapping(Instance(str), tAll, Instance(hdict))

        @param tree: PNML representation of the type
        @type tree: `snakes.pnml.Tree`
        @return: the loaded type
        @rtype: `Mapping`
        """
        return cls(tree.child("keys").child().to_obj(),
                   tree.child("items").child().to_obj(),
                   tree.child("container").child().to_obj())

class Range (Type) :
    """A type whose values are in a given range

    Notice that ranges are not built into the memory so that huge
    values can be used.

    >>> 3 in Range(1, 2**128, 2)
    True
    >>> 4 in Range(1, 2**128, 2)
    False
    """
    def __init__ (self, first, last, step=1) :
        """The values are those that the builtin `range(first, last,
        step)` would return.

        >>> Range(1, 10)
        Range(1, 10)
        >>> Range(1, 10, 2)
        Range(1, 10, 2)

        @param first: first element in the range
        @type first: `int`
        @param last: upper bound of the range, not belonging to it
        @type last: `int`
        @param step: step between elements in the range
        @type step: `int`
        """
        self._first, self._last, self._step = first, last, step
    # apidoc stop
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        >>> 1 in Range(1, 10, 2)
        True
        >>> 2 in Range(1, 10, 2)
        False
        >>> 9 in Range(1, 10, 2)
        True
        >>> 10 in Range(1, 10, 2)
        False

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        return ((self._first <= value < self._last)
                and ((value - self._first) % self._step == 0))
    def __repr__ (self) :
        """Return a string representation of the type suitable for `eval`

        >>> repr(Range(1, 2**128, 2))
        'Range(1, 340282366920938463463374607431768211456, 2)'

        @return: precise string representation
        @rtype: `str`
        """
        if self._step == 1 :
            return "Range(%s, %s)" % (self._first, self._last)
        else :
            return "Range(%s, %s, %s)" % (self._first, self._last, self._step)
    def __iter__ (self) :
        """Iterate over the elements of the type

        >>> list(iter(Range(1, 10, 3)))
        [1, 4, 7]

        @return: an iterator over the values belonging to the range
        @rtype: `generator`
        """
        return iter(xrange(self._first, self._last, self._step))
    __pnmltype__ = "range"
    def __pnmldump__ (self) :
        """Dump type to a PNML tree

        >>> Range(1, 10, 2).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="range">
          <first>
           <object type="int">
            1
           </object>
          </first>
          <last>
           <object type="int">
            10
           </object>
          </last>
          <step>
           <object type="int">
            2
           </object>
          </step>
         </type>
        </pnml>

        @return: PNML representation of the type
        @rtype: `snakes.pnml.Tree`
        """
        return Tree(self.__pnmltag__, None,
                    Tree("first", None, Tree.from_obj(self._first)),
                    Tree("last", None, Tree.from_obj(self._last)),
                    Tree("step", None, Tree.from_obj(self._step)),
                    domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Build type from its PNML representation

        >>> Range.__pnmlload__(Range(1, 10, 2).__pnmldump__())
        Range(1, 10, 2)

        @param tree: PNML tree to load
        @type tree: `snakes.pnml.Tree`
        @return: the loaded type
        @rtype: `Range`
        """
        return cls(tree.child("first").child().to_obj(),
                   tree.child("last").child().to_obj(),
                   tree.child("step").child().to_obj())

class Greater (Type) :
    """A type whose values are greater than a minimum.

    The minimum and the checked values can be of any type as soon as
    they can be compared with `>`.

    >>> 6 in Greater(3)
    True
    >>> 3 in Greater(3)
    False
    """
    def __init__ (self, min) :
        """Initialises the type

        >>> Greater(5)
        Greater(5)

        @param min: the greatest value not included in the type
        @type min: `object`
        """
        self._min = min
    # apidoc stop
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        >>> 5 in Greater(3)
        True
        >>> 5 in Greater(3.0)
        True
        >>> 3 in Greater(3.0)
        False
        >>> 1.0 in Greater(5)
        False

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        try :
            return value > self._min
        except :
            return False
    def __repr__ (self) :
        """Return a string representation of the type suitable for `eval`

        >>> repr(Greater(3))
        'Greater(3)'

        @return: precise string representation
        @rtype: `str`
        """
        return "Greater(%s)" % repr(self._min)
    __pnmltype__ = "greater"
    def __pnmldump__ (self) :
        """Dump type to its PNML representation

        >>> Greater(42).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="greater">
          <object type="int">
           42
          </object>
         </type>
        </pnml>

        @return: PNML representation of the type
        @rtype: `snakes.pnml.Tree`
        """
        return Tree(self.__pnmltag__, None,
                    Tree.from_obj(self._min),
                    domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Build type from its PNLM representation

        >>> Greater.__pnmlload__(Greater(42).__pnmldump__())
        Greater(42)

        @param tree: PNML representation to load
        @type tree: `snakes.pnml.Tree`
        @return: loaded type
        @rtype: `Greater`
        """
        return cls(tree.child().to_obj())

class GreaterOrEqual (Type) :
    """A type whose values are greater or equal than a minimum.

    See the description of `Greater`
    """
    def __init__ (self, min) :
        """Initialises the type

        >>> GreaterOrEqual(5)
        GreaterOrEqual(5)

        @param min: the minimal allowed value
        @type min: `object`
        """
        self._min = min
    # apidoc stop
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        >>> 5 in GreaterOrEqual(3)
        True
        >>> 5 in GreaterOrEqual(3.0)
        True
        >>> 3 in GreaterOrEqual(3.0)
        True
        >>> 1.0 in GreaterOrEqual(5)
        False

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        try :
            return value >= self._min
        except :
            False
    def __repr__ (self) :
        """Return a strign representation of the type suitable for `eval`

        >>> repr(GreaterOrEqual(3))
        'GreaterOrEqual(3)'

        @return: precise string representation
        @rtype: `str`
        """
        return "GreaterOrEqual(%s)" % repr(self._min)
    __pnmltype__ = "greatereq"
    def __pnmldump__ (self) :
        """Dump type to its PNML representation

        >>> GreaterOrEqual(42).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="greatereq">
          <object type="int">
           42
          </object>
         </type>
        </pnml>

        @return: PNML representation of the type
        @rtype: `snakes.pnml.Tree`
        """
        return Tree(self.__pnmltag__, None,
                    Tree.from_obj(self._min),
                    domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Build type from its PNLM representation

        >>> GreaterOrEqual.__pnmlload__(GreaterOrEqual(42).__pnmldump__())
        GreaterOrEqual(42)

        @param tree: PNML representation to load
        @type tree: `snakes.pnml.Tree`
        @return: loaded type
        @rtype: `GreaterOrEqual`
        """
        return cls(tree.child().to_obj())

class Less (Type) :
    """A type whose values are less than a maximum.

    See the description of `Greater`
    """
    def __init__ (self, max) :
        """Initialises the type

        >>> Less(5)
        Less(5)

        @param min: the smallest value not included in the type
        @type min: `object`
        """
        self._max = max
    # apidoc stop
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        >>> 5.0 in Less(5)
        False
        >>> 4.9 in Less(5)
        True
        >>> 4 in Less(5.0)
        True

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        return value < self._max
    def __repr__ (self) :
        """Return a string representation of the type suitable for `eval`

        >>> repr(Less(3))
        'Less(3)'

        @return: precise string representation
        @rtype: `str`
        """
        return "Less(%s)" % repr(self._max)
    __pnmltype__ = "less"
    def __pnmldump__ (self) :
        """Dump type to its PNML representation

        >>> Less(3).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="less">
          <object type="int">
           3
          </object>
         </type>
        </pnml>

        @return: PNML representation of the type
        @rtype: `snakes.pnml.Tree`
        """
        return Tree(self.__pnmltag__, None,
                    Tree.from_obj(self._max),
                    domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Build type from its PNML representation

        >>> Less.__pnmlload__(Less(3).__pnmldump__())
        Less(3)

        @param tree: PNML tree to load
        @type tree: `snakes.pnml.Tree`
        @return: loaded type
        @rtype: `Less`
        """
        return cls(tree.child().to_obj())

class LessOrEqual (Type) :
    """A type whose values are less than or equal to a maximum.

    See the description of `Greater`
    """
    def __init__ (self, max) :
        """Initialises the type

        >>> LessOrEqual(5)
        LessOrEqual(5)

        @param min: the greatest value the type
        @type min: `object`
        """

        self._max = max
    # apidoc stop
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        >>> 5 in LessOrEqual(5.0)
        True
        >>> 5.1 in LessOrEqual(5)
        False
        >>> 1.0 in LessOrEqual(5)
        True

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        return value <= self._max
    def __repr__ (self) :
        """Return a string representation of the type suitable for `eval`

        >>> repr(LessOrEqual(3))
        'LessOrEqual(3)'

        @return: precise string representation
        @rtype: `str`
        """
        return "LessOrEqual(%s)" % repr(self._max)
    __pnmltype__ = "lesseq"
    def __pnmldump__ (self) :
        """Dump type to its PNML representation

        >>> LessOrEqual(4).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="lesseq">
          <object type="int">
           4
          </object>
         </type>
        </pnml>

        @return: PNML representation of the type
        @rtype: `snakes.pnml.Tree`
        """
        return Tree(self.__pnmltag__, None,
                    Tree.from_obj(self._max),
                    domain=self.__pnmltype__)
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Build type from its PNML representation

        >>> LessOrEqual.__pnmlload__(LessOrEqual(4).__pnmldump__())
        LessOrEqual(4)

        @param tree: PNML tree to load
        @type tree: `snakes.pnml.Tree`
        @return: loaded type
        @rtype: `LessOrEqual`
        """
        return cls(tree.child().to_obj())

class CrossProduct (Type) :
    """A type whose values are tuples, each component of them being in
    given types. The resulting type is the cartesian cross product of
    the compound types.

    >>> (1, 3, 5) in CrossProduct(Range(1, 10), Range(1, 10, 2), Range(1, 10))
    True
    >>> (2, 4, 6) in CrossProduct(Range(1, 10), Range(1, 10, 2), Range(1, 10))
    False
    """
    def __init__ (self, *types) :
        """Initialise the type

        >>> CrossProduct(Instance(int), Instance(float))
        CrossProduct(Instance(int), Instance(float))

        @param types: the types of each component of the allowed
            tuples
        @type types: `Type`
        """
        self._types = types
        _iterable(self, *types)
    # apidoc stop
    def __repr__ (self) :
        """Return a string representation of the type suitable for `eval`

        >>> repr(CrossProduct(Range(1, 10), Range(1, 10, 2), Range(1, 10)))
        'CrossProduct(Range(1, 10), Range(1, 10, 2), Range(1, 10))'

        @return: precise string representation
        @rtype: `str`
        """
        return "CrossProduct(%s)" % ", ".join([repr(t) for t in self._types])
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        >>> (1, 3, 5) in CrossProduct(Range(1, 10), Range(1, 10, 2), Range(1, 10))
        True
        >>> (2, 4, 6) in CrossProduct(Range(1, 10), Range(1, 10, 2), Range(1, 10))
        False

        @param value: the value to check
        @type value: `object`
        @return: `True` if `value` is in the type, `False` otherwise
        @rtype: `bool`
        """
        if not isinstance(value, tuple) :
            return False
        elif len(self._types) != len(value) :
            return False
        for item, t in zip(value, self._types) :
            if not item in t :
                return False
        return True
    def __iter__ (self) :
        """A cross product is iterable if so are all its components.

        >>> list(iter(CrossProduct(Range(1, 3), Range(3, 5))))
        [(1, 3), (1, 4), (2, 3), (2, 4)]
        >>> iter(CrossProduct(Range(1, 100), tAll))
        Traceback (most recent call last):
        ...
        ValueError: iteration over non-sequence

        @return: an iterator over the values in the type
        @rtype: `generator`
        @raise ValueError: when one component is not iterable
        """
        self.__iterable__()
        return cross(self._types)
    __pnmltype__ = "crossproduct"
    def __pnmldump__ (self) :
        """Dumps type to its PNML representation

        >>> CrossProduct(Instance(int), Instance(float)).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <type domain="crossproduct">
          <type domain="instance">
           <object name="int" type="class"/>
          </type>
          <type domain="instance">
           <object name="float" type="class"/>
          </type>
         </type>
        </pnml>

        @return: PNML representation of the type
        @rtype: `snakes.pnml.Tree`
        """
        return Tree(self.__pnmltag__, None,
                    *(Tree.from_obj(t) for t in self._types),
                    **dict(domain=self.__pnmltype__))
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Build type from its PNML representation

        >>> t = CrossProduct(Instance(int), Instance(float)).__pnmldump__()
        >>> CrossProduct.__pnmlload__(t)
        CrossProduct(Instance(int), Instance(float))

        @param tree: PNML tree to load
        @type tree: `snakes.pnml.Tree`
        @return: loaded type
        @rtype: `CrossProduct`
        """
        return cls(*(child.to_obj() for child in tree.children))

tAll        = _All()
tNothing    = _Nothing()
tString     = Instance(str)
tList       = List(tAll)
tInteger    = Instance(int)
tNatural    = tInteger & GreaterOrEqual(0)
tPositive   = tInteger & Greater(0)
tFloat      = Instance(float)
tNumber     = tInteger|tFloat
tDict       = Instance(dict)
tNone       = OneOf(None)
tBoolean    = OneOf(True, False)
tTuple      = Tuple(tAll)
tPair       = Tuple(tAll, min=2, max=2)



"""A dirty trick to allow side effects in expressions

Assignement takes place when method `bind` of the expression is
called. Assigned variables are then stored in the `Substitution`
passed to `bind`. This is useful in some cases when you want to
repeat a computation in several output arcs: you do it once in the
guard and then use the bounded variable on the output arcs.

>>> import snakes.plugins
>>> snakes.plugins.load('let', 'snakes.nets', 'snk')
<module ...>
>>> from snk import *
>>> n = PetriNet('n')
>>> t = Transition('t', Expression('x is not None and let("egg, spam = iter(str(foo))", foo=42, __raise__=True)'))
>>> n.add_transition(t)
>>> n.add_place(Place('p', [dot]))
>>> n.add_input('p', 't', Variable('x'))
>>> t.modes() == [Substitution(x=dot, foo=42, egg='4', spam='2')]
True

@param args: a list of `name=value` to assign
@return: `True` if assignment could be made, `False` otherwise
@rtype: `bool`
@warning: use with care, sides effects are nasty tricks, you may
  get unexpected results while playing with `let`
"""

import inspect
import snakes.plugins

from snakes.lang.python.parser import ast, parse
from snakes.lang import unparse

class DropLet (ast.NodeTransformer) :
    def __init__ (self, names) :
        ast.NodeTransformer.__init__(self)
        self.names = set(names)
        self.calls = []
    def visit_Call (self, node) :
        func = node.func
        if func.__class__.__name__ == "Name" and func.id in self.names :
            self.calls.append((func.id,
                               node.args and node.args[0].s or None,
                               [(k.arg, unparse(k.value))
                                for k in node.keywords]))
            return ast.Name(id="True")
        return node

class DropTrue (ast.NodeTransformer) :
    def visit_BoolOp (self, node) :
        if node.op.__class__.__name__ == "And" :
            values = [self.visit(v) for v in node.values
                      if v.__class__.__name__ != "Name" or v.id != "True"]
            if values :
                node.values[:] = values
                return node
            else :
                return ast.Name(id="True")
        else :
            return self.visit(node)

def unlet (expr, *names) :
    if not names :
        names = ["let"]
    drop = DropLet(names)
    new = DropTrue().visit(drop.visit(parse(expr)))
    return unparse(new), drop.calls

class MakeLet (object) :
    def __init__ (self, globals) :
        self.globals = globals
    def match (self, match, binding) :
        env = dict(binding)
        env.update(iter(self.globals))
        exec("", env)
        old = set(env)
        exec(match, env)
        for name in set(env) - old :
            binding[name] = env[name]
    def __call__ (self, __match__=None, __raise__=False, **args) :
        try :
            __binding__ = inspect.stack()[1][0].f_locals["__binding__"]
            for name, value in args.items() :
                __binding__[name] = value
            if __match__ :
                self.match(__match__, __binding__)
        except :
            if __raise__ :
                raise
            return False
        return True

@snakes.plugins.plugin("snakes.nets")
def extend (module) :
    class PetriNet (module.PetriNet) :
        def __init__ (self, name, **args) :
            module.PetriNet.__init__(self, name, **args)
            self.globals["let"] = MakeLet(self.globals)
    return PetriNet, unlet



from snakes.lang.abcd.parser import ast

class NodeCopier (ast.NodeTransformer) :
    def copy (self, node, **replace) :
        args = {}
        for name in node._fields + node._attributes :
            old = getattr(node, name, None)
            if name in replace :
                new = replace[name]
            elif isinstance(old, list):
                new = []
                for val in old :
                    if isinstance(val, ast.AST) :
                        new.append(self.visit(val))
                    else :
                        new.append(val)
            elif isinstance(old, ast.AST):
                new = self.visit(old)
            else :
                new = old
            args[name] = new
        if hasattr(node, "st") :
            args["st"] = node.st
        return node.__class__(**args)
    def generic_visit (self, node) :
        return self.copy(node)

class ArgsBinder (NodeCopier) :
    def __init__ (self, args, buffers, nets, tasks) :
        NodeCopier.__init__(self)
        self.args = args
        self.buffers = buffers
        self.nets = nets
        self.tasks = tasks
    def visit_Name (self, node) :
        if node.id in self.args :
            return self.copy(self.args[node.id])
        else :
            return self.copy(node)
    def visit_Instance (self, node) :
        if node.net in self.nets :
            return self.copy(node, net=self.nets[node.net])
        else :
            return self.copy(node)
    def _visit_access (self, node) :
        if node.buffer in self.buffers :
            return self.copy(node, buffer=self.buffers[node.buffer])
        else :
            return self.copy(node)
    def visit_SimpleAccess (self, node) :
        return self._visit_access(node)
    def visit_FlushAccess (self, node) :
        return self._visit_access(node)
    def visit_SwapAccess (self, node) :
        return self._visit_access(node)
    def _visit_task (self, node) :
        if node.net in self.tasks :
            return self.copy(node, net=self.tasks[node.net])
        else :
            return self.copy(node)
    def visit_Spawn (self, node) :
        return self._visit_task(node)
    def visit_Wait (self, node) :
        return self._visit_task(node)
    def visit_Suspend (self, node) :
        return self._visit_task(node)
    def visit_Resume (self, node) :
        return self._visit_task(node)
    def visit_AbcdNet (self, node) :
        args = self.args.copy()
        buffers = self.buffers.copy()
        nets = self.nets.copy()
        tasks = self.tasks.copy()
        netargs = ([a.arg for a in node.args.args + node.args.kwonlyargs]
                   + [node.args.vararg, node.args.kwarg])
        copy = True
        for a in netargs :
            for d in (args, buffers, nets, tasks) :
                if a in d :
                    del d[a]
                    copy = False
        if copy :
            return self.copy(node)
        else :
            return self.__class__(args, buffers, nets, tasks).visit(node)

if __name__ == "__main__" :
    import doctest
    doctest.testmod()



"""Backport of Python 2.6 `ast` module.

AST classes are wrappers around classes in Python 2.5 `_ast` module.
The rest is copied from Python 2.6 `ast.py`.
"""

import inspect
import _ast
from _ast import PyCF_ONLY_AST
from _ast import AST

for name, cls in inspect.getmembers(_ast, inspect.isclass) :
    if issubclass(cls, AST) :
        if cls is AST :
            continue
        class _Ast (cls, AST) :
            def __init__ (self, *larg, **karg) :
                if len(larg) > 0 and len(larg) != len(self._fields) :
                    raise TypeError, ("%s constructor takes either 0 or "
                                      "%u positional arguments"
                                      % (self.__class__.__name__,
                                         len(self._fields)))
                for name, arg in zip(self._fields, larg) + karg.items() :
                    if name in self._fields :
                        setattr(self, name, arg)
        try :
            _Ast._fields = tuple(cls._fields)
        except :
            _Ast._fields = ()
        _Ast.__name__ = name
        globals()[name] = _Ast

del _Ast, cls, name

def literal_eval(node_or_string):
    """Safely evaluate an expression node or a string containing a
    Python expression. The string or node provided may only consist of
    the following Python literal structures: strings, numbers, tuples,
    lists, dicts, booleans, and `None`.
    """
    _safe_names = {'None': None, 'True': True, 'False': False}
    if isinstance(node_or_string, basestring):
        node_or_string = parse(node_or_string, mode='eval')
    if isinstance(node_or_string, Expression):
        node_or_string = node_or_string.body
    def _convert(node):
        if isinstance(node, Str):
            return node.s
        elif isinstance(node, Num):
            return node.n
        elif isinstance(node, Tuple):
            return tuple(map(_convert, node.elts))
        elif isinstance(node, List):
            return list(map(_convert, node.elts))
        elif isinstance(node, Dict):
            return dict((_convert(k), _convert(v)) for k, v
                        in zip(node.keys, node.values))
        elif isinstance(node, Name):
            if node.id in _safe_names:
                return _safe_names[node.id]
        raise ValueError('malformed string')
    return _convert(node_or_string)

def dump(node, annotate_fields=True, include_attributes=False):
    """Return a formatted dump of the tree in `node`. This is mainly
    useful for debugging purposes. The returned string will show the
    names and the values for fields. This makes the code impossible to
    evaluate, so if evaluation is wanted `annotate_fields` must be set
    to `False`. Attributes such as line numbers and column offsets are
    not dumped by default. If this is wanted, `include_attributes` can
    be set to `True`.
    """
    def _format(node):
        if isinstance(node, AST):
            fields = [(a, _format(b)) for a, b in iter_fields(node)]
            rv = '%s(%s' % (node.__class__.__name__, ', '.join(
                ('%s=%s' % field for field in fields)
                if annotate_fields else
                (b for a, b in fields)
            ))
            if include_attributes and node._attributes:
                rv += fields and ', ' or ' '
                rv += ', '.join('%s=%s' % (a, _format(getattr(node, a)))
                                for a in node._attributes)
            return rv + ')'
        elif isinstance(node, list):
            return '[%s]' % ', '.join(_format(x) for x in node)
        return repr(node)
    if not isinstance(node, AST):
        raise TypeError('expected AST, got %r' % node.__class__.__name__)
    return _format(node)

def copy_location(new_node, old_node):
    """Copy source location (`lineno` and `col_offset` attributes)
    from `old_node` to `new_node` if possible, and return `new_node`.
    """
    for attr in 'lineno', 'col_offset':
        if attr in old_node._attributes and attr in new_node._attributes \
           and hasattr(old_node, attr):
            setattr(new_node, attr, getattr(old_node, attr))
    return new_node

def fix_missing_locations(node):
    """When you compile a node tree with `compile()`, the compiler
    expects `lineno` and `col_offset` attributes for every node that
    supports them. This is rather tedious to fill in for generated
    nodes, so this helper adds these attributes recursively where not
    already set, by setting them to the values of the parent node. It
    works recursively starting at `node`.
    """
    def _fix(node, lineno, col_offset):
        if 'lineno' in node._attributes:
            if not hasattr(node, 'lineno'):
                node.lineno = lineno
            else:
                lineno = node.lineno
        if 'col_offset' in node._attributes:
            if not hasattr(node, 'col_offset'):
                node.col_offset = col_offset
            else:
                col_offset = node.col_offset
        for child in iter_child_nodes(node):
            _fix(child, lineno, col_offset)
    _fix(node, 1, 0)
    return node

def increment_lineno(node, n=1):
    '''Increment the line number of each node in the tree starting at
    `node` by `n`. This is useful to "move code" to a different
    location in a file.
    '''
    if 'lineno' in node._attributes:
        node.lineno = getattr(node, 'lineno', 0) + n
    for child in walk(node):
        if 'lineno' in child._attributes:
            child.lineno = getattr(child, 'lineno', 0) + n
    return node

def iter_fields(node):
    """Yield a tuple of `(fieldname, value)` for each field in
    `node._fields` that is present on `node`.
    """
    for field in node._fields:
        try:
            yield field, getattr(node, field)
        except AttributeError:
            pass

def iter_child_nodes(node):
    """Yield all direct child nodes of `node`, that is, all fields
    that are nodes and all items of fields that are lists of nodes.
    """
    for name, field in iter_fields(node):
        if isinstance(field, AST):
            yield field
        elif isinstance(field, list):
            for item in field:
                if isinstance(item, AST):
                    yield item

def get_docstring(node, clean=True):
    """Return the docstring for the given `node` or `None` if no
    docstring can be found. If the node provided does not have
    docstrings a `TypeError` will be raised.
    """
    if not isinstance(node, (FunctionDef, ClassDef, Module)):
        raise TypeError("%r can't have docstrings" % node.__class__.__name__)
    if node.body and isinstance(node.body[0], Expr) and \
       isinstance(node.body[0].value, Str):
        if clean:
            return inspect.cleandoc(node.body[0].value.s)
        return node.body[0].value.s

def walk(node):
    """Recursively yield all child nodes of `node`, in no specified
    order. This is useful if you only want to modify nodes in place
    and don't care about the context.
    """
    from collections import deque
    todo = deque([node])
    while todo:
        node = todo.popleft()
        todo.extend(iter_child_nodes(node))
        yield node

class NodeVisitor(object):
    """A node visitor base class that walks the abstract syntax tree
    and calls a visitor function for every node found. This function
    may return a value which is forwarded by the `visit` method.

    This class is meant to be subclassed, with the subclass adding
    visitor methods.

    Per default the visitor functions for the nodes are `visit_` +
    class name of the node. So a `TryFinally` node visit function
    would be `visit_TryFinally`. This behavior can be changed by
    overriding the `visit` method. If no visitor function exists for a
    node (return value `None`) the `generic_visit` visitor is used
    instead.

    Don't use the `NodeVisitor` if you want to apply changes to nodes
    during traversing. For this a special visitor exists
    (`NodeTransformer`) that allows modifications.
    """
    def visit(self, node):
        """Visit a node."""
        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)
    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""
        for field, value in iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, AST):
                        self.visit(item)
            elif isinstance(value, AST):
                self.visit(value)

class NodeTransformer(NodeVisitor):
    """A `NodeVisitor` subclass that walks the abstract syntax tree
    and allows modification of nodes.

    The `NodeTransformer` will walk the AST and use the return value
    of the visitor methods to replace or remove the old node. If the
    return value of the visitor method is `None`, the node will be
    removed from its location, otherwise it is replaced with the
    return value. The return value may be the original node in which
    case no replacement takes place.

    Here is an example transformer that rewrites all occurrences of
    name lookups (`foo`) to `data['foo']`:

        class RewriteName(NodeTransformer):
            def visit_Name(self, node):
                return copy_location(Subscript(
                    value=Name(id='data', ctx=Load()),
                    slice=Index(value=Str(s=node.id)),
                    ctx=node.ctx
                   ), node)

    Keep in mind that if the node you're operating on has child
    nodes you must either transform the child nodes yourself or call
    the `generic_visit` method for the node first.

    For nodes that were part of a collection of statements (that
    applies to all statement nodes), the visitor may also return a
    list of nodes rather than just a single node.

    Usually you use the transformer like this:

        node = YourTransformer().visit(node)

    """
    def generic_visit(self, node):
        for field, old_value in iter_fields(node):
            old_value = getattr(node, field, None)
            if isinstance(old_value, list):
                new_values = []
                for value in old_value:
                    if isinstance(value, AST):
                        value = self.visit(value)
                        if value is None:
                            continue
                        elif not isinstance(value, AST):
                            new_values.extend(value)
                            continue
                    new_values.append(value)
                old_value[:] = new_values
            elif isinstance(old_value, AST):
                new_node = self.visit(old_value)
                if new_node is None:
                    delattr(node, field)
                else:
                    setattr(node, field, new_node)
        return node

def _ast2ast (node) :
    new = globals()[node.__class__.__name__]()
    if not hasattr(node, "_fields") or node._fields is None :
        node._fields = ()
    for name, field in iter_fields(node) :
        new_field = field
        if field is None :
            new_field = None
        elif isinstance(field, AST) :
            new_field = _ast2ast(field)
        elif isinstance(field, list) :
            new_field = []
            for value in field :
                if isinstance(value, AST) :
                    new_field.append(_ast2ast(value))
                else :
                    new_field.append(value)
        setattr(new, name, new_field)
    copy_location(new, node)
    return new

def parse(expr, filename='<unknown>', mode='exec'):
    """Parse an expression into an AST node. Equivalent to
    `compile(expr, filename, mode, PyCF_ONLY_AST)`.
    """
    return _ast2ast(compile(expr, filename, mode, PyCF_ONLY_AST))



"Usage: unparse.py <path to source file>"
import sys
import _ast
from snakes.compat import *

def interleave(inter, f, seq):
    """Call f on each item in seq, calling inter() in between.
    """
    seq = iter(seq)
    try:
        f(next(seq))
    except StopIteration:
        pass
    else:
        for x in seq:
            inter()
            f(x)

class Unparser:
    """Methods in this class recursively traverse an AST and output
    source code for the abstract syntax; original formatting is
    disregarged.
    """

    def __init__(self, tree, file = sys.stdout):
        """Unparser(tree, file=sys.stdout) -> None.

        Print the source for tree to file.
        """
        self.f = file
        self._indent = 0
        self.dispatch(tree)
        self.f.write("\n")
        self.f.flush()

    def fill(self, text = ""):
        """Indent a piece of text, according to the current indentation
        level
        """
        self.f.write("\n" + "    "*self._indent + text)

    def write(self, text):
        "Append a piece of text to the current line."
        self.f.write(text)

    def enter(self):
        "Print ':', and increase the indentation."
        self.write(":")
        self._indent += 1

    def leave(self):
        "Decrease the indentation level."
        self._indent -= 1

    def dispatch(self, tree):
        "Dispatcher function, dispatching tree type T to method _T."
        if isinstance(tree, list):
            for t in tree:
                self.dispatch(t)
            return
        meth = getattr(self, "_"+tree.__class__.__name__)
        meth(tree)


    ############### Unparsing methods ######################
    # There should be one method per concrete grammar type #
    # Constructors should be grouped by sum type. Ideally, #
    # this would follow the order in the grammar, but      #
    # currently doesn't.                                   #
    ########################################################

    def _Module(self, tree):
        for stmt in tree.body:
            self.dispatch(stmt)

    # stmt
    def _Expr(self, tree):
        self.fill()
        self.dispatch(tree.value)

    def _Import(self, t):
        self.fill("import ")
        interleave(lambda: self.write(", "), self.dispatch, t.names)

    def _ImportFrom(self, t):
        self.fill("from ")
        self.write(t.module)
        self.write(" import ")
        interleave(lambda: self.write(", "), self.dispatch, t.names)
        # XXX(jpe) what is level for?

    def _Assign(self, t):
        self.fill()
        for target in t.targets:
            self.dispatch(target)
            self.write(" = ")
        self.dispatch(t.value)

    def _AugAssign(self, t):
        self.fill()
        self.dispatch(t.target)
        self.write(" "+self.binop[t.op.__class__.__name__]+"= ")
        self.dispatch(t.value)

    def _Return(self, t):
        self.fill("return")
        if t.value:
            self.write(" ")
            self.dispatch(t.value)

    def _Pass(self, t):
        self.fill("pass")

    def _Break(self, t):
        self.fill("break")

    def _Continue(self, t):
        self.fill("continue")

    def _Delete(self, t):
        self.fill("del ")
        self.dispatch(t.targets)

    def _Assert(self, t):
        self.fill("assert ")
        self.dispatch(t.test)
        if t.msg:
            self.write(", ")
            self.dispatch(t.msg)

    def _Exec(self, t):
        self.fill("exec ")
        self.dispatch(t.body)
        if t.globals:
            self.write(" in ")
            self.dispatch(t.globals)
        if t.locals:
            self.write(", ")
            self.dispatch(t.locals)

    def _Print(self, t):
        self.fill("print ")
        do_comma = False
        if t.dest:
            self.write(">>")
            self.dispatch(t.dest)
            do_comma = True
        for e in t.values:
            if do_comma:self.write(", ")
            else:do_comma=True
            self.dispatch(e)
        if not t.nl:
            self.write(",")

    def _Global(self, t):
        self.fill("global ")
        interleave(lambda: self.write(", "), self.write, t.names)

    def _Yield(self, t):
        self.write("(")
        self.write("yield")
        if t.value:
            self.write(" ")
            self.dispatch(t.value)
        self.write(")")

    def _Raise(self, t):
        self.fill('raise ')
        if t.type:
            self.dispatch(t.type)
        if t.inst:
            self.write(", ")
            self.dispatch(t.inst)
        if t.tback:
            self.write(", ")
            self.dispatch(t.tback)

    def _TryExcept(self, t):
        self.fill("try")
        self.enter()
        self.dispatch(t.body)
        self.leave()

        for ex in t.handlers:
            self.dispatch(ex)
        if t.orelse:
            self.fill("else")
            self.enter()
            self.dispatch(t.orelse)
            self.leave()

    def _TryFinally(self, t):
        self.fill("try")
        self.enter()
        self.dispatch(t.body)
        self.leave()

        self.fill("finally")
        self.enter()
        self.dispatch(t.finalbody)
        self.leave()

    def _ExceptHandler(self, t):
        self.fill("except")
        if t.type:
            self.write(" ")
            self.dispatch(t.type)
        if t.name:
            self.write(", ")
            self.dispatch(t.name)
        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _ClassDef(self, t):
        self.write("\n")
        self.fill("class "+t.name)
        if t.bases:
            self.write("(")
            for a in t.bases:
                self.dispatch(a)
                self.write(", ")
            self.write(")")
        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _FunctionDef(self, t):
        self.write("\n")
        for deco in t.decorator_list:
            self.fill("@")
            self.dispatch(deco)
        self.fill("def "+t.name + "(")
        self.dispatch(t.args)
        self.write(")")
        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _For(self, t):
        self.fill("for ")
        self.dispatch(t.target)
        self.write(" in ")
        self.dispatch(t.iter)
        self.enter()
        self.dispatch(t.body)
        self.leave()
        if t.orelse:
            self.fill("else")
            self.enter()
            self.dispatch(t.orelse)
            self.leave

    def _If(self, t):
        self.fill("if ")
        self.dispatch(t.test)
        self.enter()
        # XXX elif?
        self.dispatch(t.body)
        self.leave()
        if t.orelse:
            self.fill("else")
            self.enter()
            self.dispatch(t.orelse)
            self.leave()

    def _While(self, t):
        self.fill("while ")
        self.dispatch(t.test)
        self.enter()
        self.dispatch(t.body)
        self.leave()
        if t.orelse:
            self.fill("else")
            self.enter()
            self.dispatch(t.orelse)
            self.leave

    def _With(self, t):
        self.fill("with ")
        self.dispatch(t.context_expr)
        if t.optional_vars:
            self.write(" as ")
            self.dispatch(t.optional_vars)
        self.enter()
        self.dispatch(t.body)
        self.leave()

    # expr
    def _Str(self, tree):
        self.write(repr(tree.s))

    def _Name(self, t):
        self.write(t.id)

    def _Repr(self, t):
        self.write("`")
        self.dispatch(t.value)
        self.write("`")

    def _Num(self, t):
        self.write(repr(t.n))

    def _List(self, t):
        self.write("[")
        interleave(lambda: self.write(", "), self.dispatch, t.elts)
        self.write("]")

    def _ListComp(self, t):
        self.write("[")
        self.dispatch(t.elt)
        for gen in t.generators:
            self.dispatch(gen)
        self.write("]")

    def _GeneratorExp(self, t):
        self.write("(")
        self.dispatch(t.elt)
        for gen in t.generators:
            self.dispatch(gen)
        self.write(")")

    def _comprehension(self, t):
        self.write(" for ")
        self.dispatch(t.target)
        self.write(" in ")
        self.dispatch(t.iter)
        for if_clause in t.ifs:
            self.write(" if ")
            self.dispatch(if_clause)

    def _IfExp(self, t):
        self.write("(")
        self.dispatch(t.body)
        self.write(" if ")
        self.dispatch(t.test)
        self.write(" else ")
        self.dispatch(t.orelse)
        self.write(")")

    def _Dict(self, t):
        self.write("{")
        def writem(arg):
            (k, v) = arg
            self.dispatch(k)
            self.write(": ")
            self.dispatch(v)
        interleave(lambda: self.write(", "), writem, zip(t.keys, t.values))
        self.write("}")

    def _Set(self, t) :
        self.write("set([")
        if len(t.elts) == 1:
            (elt,) = t.elts
            self.dispatch(elt)
            self.write(",")
        else:
            interleave(lambda: self.write(", "), self.dispatch, t.elts)
        self.write("])")

    def _Tuple(self, t):
        self.write("(")
        if len(t.elts) == 1:
            (elt,) = t.elts
            self.dispatch(elt)
            self.write(",")
        else:
            interleave(lambda: self.write(", "), self.dispatch, t.elts)
        self.write(")")

    unop = {"Invert":"~", "Not": "not", "UAdd":"+", "USub":"-"}
    def _UnaryOp(self, t):
        self.write(self.unop[t.op.__class__.__name__])
        self.write("(")
        self.dispatch(t.operand)
        self.write(")")

    binop = { "Add":"+", "Sub":"-", "Mult":"*", "Div":"/", "Mod":"%",
              "LShift":">>", "RShift":"<<", "BitOr":"|", "BitXor":"^", "BitAnd":"&",
              "FloorDiv":"//", "Pow": "**"}
    def _BinOp(self, t):
        self.write("(")
        self.dispatch(t.left)
        self.write(" " + self.binop[t.op.__class__.__name__] + " ")
        self.dispatch(t.right)
        self.write(")")

    cmpops = {"Eq":"==", "NotEq":"!=", "Lt":"<", "LtE":"<=", "Gt":">", "GtE":">=",
              "Is":"is", "IsNot":"is not", "In":"in", "NotIn":"not in"}
    def _Compare(self, t):
        self.write("(")
        self.dispatch(t.left)
        for o, e in zip(t.ops, t.comparators):
            self.write(" " + self.cmpops[o.__class__.__name__] + " ")
            self.dispatch(e)
        self.write(")")

    boolops = {"And": 'and', "Or": 'or'}
    def _BoolOp(self, t):
        self.write("(")
        s = " %s " % self.boolops[t.op.__class__.__name__]
        interleave(lambda: self.write(s), self.dispatch, t.values)
        self.write(")")

    def _Attribute(self,t):
        self.dispatch(t.value)
        self.write(".")
        self.write(t.attr)

    def _Call(self, t):
        self.dispatch(t.func)
        self.write("(")
        comma = False
        for e in t.args:
            if comma: self.write(", ")
            else: comma = True
            self.dispatch(e)
        for e in t.keywords:
            if comma: self.write(", ")
            else: comma = True
            self.dispatch(e)
        if t.starargs:
            if comma: self.write(", ")
            else: comma = True
            self.write("*")
            self.dispatch(t.starargs)
        if t.kwargs:
            if comma: self.write(", ")
            else: comma = True
            self.write("**")
            self.dispatch(t.kwargs)
        self.write(")")

    def _Subscript(self, t):
        self.dispatch(t.value)
        self.write("[")
        self.dispatch(t.slice)
        self.write("]")

    # slice
    def _Ellipsis(self, t):
        self.write("...")

    def _Index(self, t):
        self.dispatch(t.value)

    def _Slice(self, t):
        if t.lower:
            self.dispatch(t.lower)
        self.write(":")
        if t.upper:
            self.dispatch(t.upper)
        if t.step:
            self.write(":")
            self.dispatch(t.step)

    def _ExtSlice(self, t):
        interleave(lambda: self.write(', '), self.dispatch, t.dims)

    # others
    def _arguments(self, t):
        first = True
        nonDef = len(t.args)-len(t.defaults)
        for a in t.args[0:nonDef]:
            if first:first = False
            else: self.write(", ")
            self.dispatch(a)
        for a,d in zip(t.args[nonDef:], t.defaults):
            if first:first = False
            else: self.write(", ")
            self.dispatch(a),
            self.write("=")
            self.dispatch(d)
        if t.vararg:
            if first:first = False
            else: self.write(", ")
            self.write("*"+t.vararg)
        if t.kwarg:
            if first:first = False
            else: self.write(", ")
            self.write("**"+t.kwarg)

    def _keyword(self, t):
        self.write(t.arg)
        self.write("=")
        self.dispatch(t.value)

    def _Lambda(self, t):
        self.write("lambda ")
        self.dispatch(t.args)
        self.write(": ")
        self.dispatch(t.body)

    def _alias(self, t):
        self.write(t.name)
        if t.asname:
            self.write(" as "+t.asname)

def roundtrip(filename, output=sys.stdout):
    source = open(filename).read()
    tree = compile(source, filename, "exec", _ast.PyCF_ONLY_AST)
    Unparser(tree, output)



def testdir(a):
    try:
        names = [n for n in os.listdir(a) if n.endswith('.py')]
    except OSError:
        sys.stderr.write("Directory not readable: %s\n" % a)
    else:
        for n in names:
            fullname = os.path.join(a, n)
            if os.path.isfile(fullname):
                output = io.StringIO()
                print('Testing %s' % fullname)
                try:
                    roundtrip(fullname, output)
                except Exception:
                    e = sys.exc_info()[1]
                    print('  Failed to compile, exception is %r' % e)
            elif os.path.isdir(fullname):
                testdir(fullname)

def main(args):
    if args[0] == '--testdir':
        for a in args[1:]:
            testdir(a)
    else:
        for a in args:
            roundtrip(a)

if __name__=='__main__':
    main(sys.argv[1:])



import doctest, sys, os, glob

retcode = 0

import snakes
version = open("VERSION").read().strip()
if snakes.version != version :
    print("Mismatched versions:")
    print("  snakes.version = %r" % snakes.version)
    print("  VERSION = %r" % version)
    sys.exit(1)

def test (module) :
    print("  Testing '%s'" % module.__name__)
    f, t = doctest.testmod(module, #verbose=True,
                           optionflags=doctest.NORMALIZE_WHITESPACE
                           | doctest.REPORT_ONLY_FIRST_FAILURE
                           | doctest.ELLIPSIS)
    return f

modules = ["snakes",
           "snakes.hashables",
           "snakes.lang",
           "snakes.lang.python.parser",
           "snakes.lang.abcd.parser",
           "snakes.lang.ctlstar.parser",
           "snakes.data",
           "snakes.typing",
           "snakes.nets",
           "snakes.pnml",
           "snakes.plugins",
           "snakes.plugins.pos",
           "snakes.plugins.status",
           "snakes.plugins.ops",
           "snakes.plugins.synchro",
           "snakes.plugins.hello",
           "snakes.plugins.gv",
           "snakes.plugins.clusters",
           "snakes.plugins.labels",
           "snakes.utils.abcd.build",
           ]

stop = False
if len(sys.argv) > 1 :
    if sys.argv[1] == "--stop" :
        stop = True
        del sys.argv[1]

doscripts = True
if len(sys.argv) > 1 :
    modules = sys.argv[1:]
    doscripts = False

for modname in modules :
    try :
        __import__(modname)
        retcode = max(retcode, test(sys.modules[modname]))
        if retcode and stop :
            break
    except :
        print("  Could not test %r:" % modname)
        c, e, t = sys.exc_info()
        print("    %s: %s" % (c.__name__, e))

if doscripts :
    for script in (glob.glob("test-scripts/test*.sh")
                   + glob.glob("test-scripts/test*.py")) :
        print("  Running '%s'" % script)
        retcode = max(retcode, os.system(script))
        if retcode and stop :
            break

sys.exit(retcode)



"""An example plugin that allows instances of `PetriNet` to
say hello. The source code can be used as a starting
example."""

import snakes.plugins

@snakes.plugins.plugin("snakes.nets")
def extend (module) :
    "Extends `module`"
    class PetriNet (module.PetriNet) :
        "Extension of the class `PetriNet` in `module`"
        def __init__ (self, name, **args) :
            """Add new keyword argument `hello`

            >>> PetriNet('N').hello()
            Hello from N
            >>> PetriNet('N', hello='Hi! This is %s...').hello()
            Hi! This is N...

            @param args: plugin options
            @keyword hello: the message to print, with
                `%s` where the net name should appear.
            @type hello: `str`
            """
            self._hello = args.pop("hello", "Hello from %s")
            module.PetriNet.__init__(self, name, **args)
        def hello (self) :
            "Ask the net to say hello"
            print(self._hello % self.name)
    return PetriNet



#-*- encoding: latin-1
"""A plugin to add positions to the nodes.

`Place` and `Transition` are added an optional argument for the
constructor `pos=(x,y)` to set their position. Moreover, these classes
are added an attribute `pos` that holds a pair of numbers with
attributes `x` and `y` and methods `shift(dx, dy)` and `moveto(x, y)`.
So, when the plugin is loaded, we can specify and retreive nodes
positions:

>>> import snakes.plugins
>>> snakes.plugins.load('pos', 'snakes.nets', 'nets')
<module ...>
>>> from nets import PetriNet, Place, Transition
>>> n = PetriNet('N')
>>> n.add_place(Place('p00'))
>>> t10 = Transition('t10', pos=(1, 0))
>>> n.add_transition(t10)
>>> n.add_place(Place('p11', pos=(1, 1)))
>>> n.add_transition(Transition('t01', pos=(0, 1)))
>>> t10.pos
Position(1, 0)
>>> t10.pos.x
1
>>> t10.pos.y
0
>>> t10.pos()
(1, 0)

Nodes positions is immutable, we must use method `moveto` to change
positions:

>>> t10.pos.y = 1
Traceback (most recent call last):
  ...
AttributeError: readonly attribute
>>> t10.pos.moveto(t10.pos.x, 1)
>>> t10.pos
Position(1, 1)

Petri nets are added methods `bbox()` that returns a pair of extrema
`((xmin, ymin), (xmax, ymax))`, a method `shift(dx, dy)` that shift
all the nodes, and a method `transpose()` that rotates the net in such
a way that the top-down direction becomes left-right:

>>> n.bbox()
((0, 0), (1, 1))
>>> n.shift(1, 2)
>>> n.bbox()
((1, 2), (2, 3))
>>> n.node('t01').copy().pos
Position(1, 3)
>>> n.transpose()
>>> n.node('t01').pos
Position(-3, 1)
"""

from snakes import SnakesError
from snakes.compat import *
from snakes.plugins import plugin, new_instance
from snakes.pnml import Tree

class Position (object) :
    "The position of a node"
    def __init__ (self, x, y) :
        """Constructor expects the Cartesian coordinates of the node,
        they can be provided as `float` or `int`.

        @param x: horizontal position
        @type x: `float`
        @param y: vertical position
        @type y: `float`
        """
        self.__dict__["x"] = x
        self.__dict__["y"] = y
    # apidoc skip
    def __str__ (self) :
        return "(%s, %s)" % (str(self.x), str(self.y))
    # apidoc skip
    def __repr__ (self) :
        return "Position(%s, %s)" % (str(self.x), str(self.y))
    # apidoc skip
    def __setattr__ (self, name, value) :
        if name in ("x", "y") :
            raise AttributeError("readonly attribute")
        else :
            self.__dict__[name] = value
    def moveto (self, x, y) :
        """Change current coordinates to the specified position

        @param x: horizontal position
        @type x: `float`
        @param y: vertical position
        @type y: `float`
        """
        self.__init__(x, y)
    def shift (self, dx, dy) :
        """Shift current coordinates by the specified amount.

        @param dx: horizontal shift
        @type dx: `float`
        @param dy: vertical shift
        @type dy: `float`
        """
        self.__init__(self.x + dx, self.y + dy)
    def __getitem__ (self, index) :
        """Access coordinates by index

        >>> Position(1, 2)[0]
        1
        >>> Position(1, 2)[1]
        2
        >>> Position(1, 2)[42]
        Traceback (most recent call last):
          ...
        IndexError: Position index out of range

        @param index: 0 for `x` coordinate, 1 for `y`
        @type index: `int`
        @raise IndexError: when `index not in {0, 1}`
        """
        if index == 0 :
            return self.x
        elif index == 1 :
            return self.y
        else :
            raise IndexError("Position index out of range")
    def __iter__ (self) :
        """Successively yield `x` and `y` coordinates

        >>> list(Position(1, 2))
        [1, 2]
        """
        yield self.x
        yield self.y
    def __call__ (self) :
        """Return the position as a pair of values

        >>> Position(1, 2.0)()
        (1, 2.0)

        @return: the pair of coordinates `(x, y)`
        @rtype: `tuple`
        """
        return (self.x, self.y)

@plugin("snakes.nets")
def extend (module) :
    class Place (module.Place) :
        def __init__ (self, name, tokens=[], check=None, **args) :
            """If no position is given `(0, 0)` is chosen

            >>> Place('p').pos
            Position(0, 0)
            >>> Place('p', pos=(1,2)).pos
            Position(1, 2)

            @keyword pos: the position of the new place
            @type pos: `tuple`
            """
            x, y = args.pop("pos", (0, 0))
            self.pos = Position(x, y)
            module.Place.__init__(self, name, tokens, check, **args)
        # apidoc skip
        def copy (self, name=None, **args) :
            x, y = args.pop("pos", self.pos())
            result = module.Place.copy(self, name, **args)
            result.pos.moveto(x, y)
            return result
        # apidoc skip
        def __pnmldump__ (self) :
            """
            >>> p = Place('p', pos=(1, 2))
            >>> p.__pnmldump__()
            <?xml version="1.0" encoding="utf-8"?>
            <pnml>
             <place id="p">
              <type domain="universal"/>
              <initialMarking>
               <multiset/>
              </initialMarking>
              <graphics>
               <position x="1" y="2"/>
              </graphics>
             </place>
            </pnml>
            """
            t = module.Place.__pnmldump__(self)
            try :
                gfx = t.child("graphics")
            except SnakesError :
                gfx = Tree("graphics", None)
                t.add_child(gfx)
            gfx.add_child(Tree("position", None,
                               x=str(self.pos.x),
                               y=str(self.pos.y)))
            return t
        # apidoc skip
        @classmethod
        def __pnmlload__ (cls, tree) :
            """
            >>> old = Place('p', pos=(1, 2))
            >>> p = old.__pnmldump__()
            >>> new = Place.__pnmlload__(p)
            >>> new.pos
            Position(1, 2)
            >>> new
            Place('p', MultiSet([]), tAll)
            >>> new.__class__
            <class 'snakes.plugins.pos.Place'>
            """
            result = new_instance(cls, module.Place.__pnmlload__(tree))
            try :
                p = tree.child("graphics").child("position")
                x, y = eval(p["x"]), eval(p["y"])
                result.pos = Position(x, y)
            except SnakesError :
                result.pos = Position(0, 0)
            return result
    class Transition (module.Transition) :
        def __init__ (self, name, guard=None, **args) :
            """If no position is given `(0, 0)` is chosen

            >>> Transition('t').pos
            Position(0, 0)
            >>> Transition('t', pos=(1,2)).pos
            Position(1, 2)

            @keyword pos: the position of the new transition
            @type pos: `tuple`
            """
            x, y = args.pop("pos", (0, 0))
            self.pos = Position(x, y)
            module.Transition.__init__(self, name, guard, **args)
        # apidoc skip
        def copy (self, name=None, **args) :
            x, y = args.pop("pos", self.pos())
            result = module.Transition.copy(self, name, **args)
            result.pos.moveto(x, y)
            return result
        # apidoc skip
        def __pnmldump__ (self) :
            """
            >>> t = Transition('t', pos=(2, 1))
            >>> t.__pnmldump__()
            <?xml version="1.0" encoding="utf-8"?>
            <pnml>
             <transition id="t">
              <graphics>
               <position x="2" y="1"/>
              </graphics>
             </transition>
            </pnml>
            """
            t = module.Transition.__pnmldump__(self)
            t.add_child(Tree("graphics", None,
                             Tree("position", None,
                                  x=str(self.pos.x),
                                  y=str(self.pos.y))))
            return t
        # apidoc skip
        @classmethod
        def __pnmlload__ (cls, tree) :
            """
            >>> old = Transition('t', pos=(2, 1))
            >>> p = old.__pnmldump__()
            >>> new = Transition.__pnmlload__(p)
            >>> new.pos
            Position(2, 1)
            >>> new
            Transition('t', Expression('True'))
            >>> new.__class__
            <class 'snakes.plugins.pos.Transition'>
            """
            result = new_instance(cls, module.Transition.__pnmlload__(tree))
            try :
                p = tree.child("graphics").child("position")
                x, y = eval(p["x"]), eval(p["y"])
                result.pos = Position(x, y)
            except SnakesError :
                result.pos = Position(0, 0)
            return result
    class PetriNet (module.PetriNet) :
        def add_place (self, place, **args) :
            """Position can be set also when a place is added to the
            net.

            >>> n = PetriNet('n')
            >>> n.add_place(Place('a', pos=(1, 2)))
            >>> n.node('a').pos
            Position(1, 2)
            >>> n.add_place(Place('b'), pos=(3,4))
            >>> n.node('b').pos
            Position(3, 4)
            >>> n.add_place(Place('c', pos=(42, 42)), pos=(5, 6))
            >>> n.node('c').pos
            Position(5, 6)

            @keyword pos: the position of the added place
            @type pos: `tuple`
            """
            if "pos" in args :
                x, y = args.pop("pos")
                place.pos.moveto(x, y)
            module.PetriNet.add_place(self, place, **args)
        def add_transition (self, trans, **args) :
            """Position can be set also when a transitions is added to
            the net. See method `add_place` above.

            @keyword pos: the position of the added transition
            @type pos: `tuple`
            """
            if "pos" in args :
                x, y = args.pop("pos")
                trans.pos.moveto(x, y)
            module.PetriNet.add_transition(self, trans, **args)
        def merge_places (self, target, sources, **args) :
            """When places are merged, the position of the new place
            is the barycentre of the positions of the merged nodes.
            Optionally, position can be specified in the call the
            `merge_places`.

            >>> n = PetriNet('n')
            >>> n.add_place(Place('a', pos=(1,2)))
            >>> n.add_place(Place('b', pos=(3,4)))
            >>> n.merge_places('c', ['a', 'b'])
            >>> n.node('c').pos
            Position(2.0, 3.0)
            >>> n.merge_places('d', ['a', 'b'], pos=(0, 0))
            >>> n.node('d').pos
            Position(0, 0)

            @keyword pos: the position of the added transition
            @type pos: `tuple`
            """
            pos = args.pop("pos", None)
            module.PetriNet.merge_places(self, target, sources, **args)
            if pos is None :
                pos = reduce(complex.__add__,
                             (complex(*self._place[name].pos())
                              for name in sources)) / len(sources)
                x, y = pos.real, pos.imag
            else :
                x, y = pos
            self._place[target].pos.moveto(x, y)
        def merge_transitions (self, target, sources, **args) :
            """See method `merge_places` above.
            """
            pos = args.pop("pos", None)
            module.PetriNet.merge_transitions(self, target, sources, **args)
            if pos is None :
                pos = reduce(complex.__add__,
                             (complex(*self._trans[name].pos())
                              for name in sources)) / len(sources)
                x, y = pos.real, pos.imag
            else :
                x, y = pos
            self._trans[target].pos.moveto(x, y)
        def bbox (self) :
            """The bounding box of the net, that is, the smallest
            rectangle that contains all nodes coordinates.

            @return: rectangle coordinates as `((xmin, ymin), (xmax,
                ymax))`
            @rtype: tuple
            """
            if len(self._node) == 0 :
                return (0, 0), (0, 0)
            else :
                nodes = iter(self._node.values())
                xmin, ymin = next(nodes).pos()
                xmax, ymax = xmin, ymin
                for n in nodes :
                    x, y = n.pos()
                    xmin = min(xmin, x)
                    xmax = max(xmax, x)
                    ymin = min(ymin, y)
                    ymax = max(ymax, y)
                return (xmin, ymin), (xmax, ymax)
        def shift (self, dx, dy) :
            """Shift every node by `(dx, dy)`

            @param dx: horizontal shift
            @type dx: `float`
            @param dy: vertical shift
            @type dy: `float`
            """
            for node in self.node() :
                node.pos.shift(dx, dy)
        def transpose (self) :
            """Perform a clockwise 90° rotation of node coordinates,
            ie, change every position `(x, y)` to `(-y, x)`
            """
            for node in self.node() :
                x, y = node.pos()
                node.pos.moveto(-y, x)
    return Place, Transition, PetriNet, Position
